from peft import PeftModel, prepare_model_for_kbit_training
from transformers import BitsAndBytesConfig
import torch

BITS_AND_BYTES_CONFIG_4BIT = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_use_double_quant=True,
    bnb_4bit_quant_type="nf4",
    bnb_4bit_compute_dtype=torch.bfloat16
)

def load_peft_model_with_adpaters(model, adapters):
  model = PeftModel.from_pretrained(
      model, 
      adapters
  )
  model.gradient_checkouting_enabled()
  model = prepare_model_for_kbit_training(model)

  return model 


def print_trainable_parameters(model):
    """
    Prints the number of trainable parameters in the model.
    """
    trainable_params = 0
    all_param = 0
    for _, param in model.named_parameters():
        all_param += param.numel()
        if param.requires_grad:
            trainable_params += param.numel()
    print(
        f"trainable params: {trainable_params} || all params: {all_param} || trainable%: {100 * trainable_params / all_param}"
    )


from pytesseract import image_to_string
from PIL import Image
from io import BytesIO
import pypdfium2

def convert_pdf_to_images(file_path, scale=300/72):
  pdf_file = pypdfium2.PdfDocument(file_path)

  page_indices = [i for i in range(len(pdf_file))]

  renderer = pdf_file.render(
    pypdfium2.PdfBitmap.to_pil,
    page_indices=page_indices,
    scale=scale,
  )

  final_images = []

  for i, image in zip(page_indices, renderer):
    image_byte_array = BytesIO()
    image.save(image_byte_array, format='jpeg', optimize=True)
    image_byte_array = image_byte_array.getvalue()
    final_images.append(dict({i: image_byte_array}))

  return final_images


def extract_text_from_imgs(list_dict_final_images):
  image_list = [list(data.values())[0] for data in list_dict_final_images]
  image_content = []
  
  for index, image_bytes in enumerate(image_list):
    image = Image.open(BytesIO(image_bytes))
    raw_text = str(image_to_string(image))
    image_content.append(raw_text)

  return "\n".join(image_content)


def extract_text_from_pdf(url: str):
  images_list = convert_pdf_to_images(url)
  text_with_pytesseract = extract_text_from_imgs(images_list)

  return text_with_pytesseract


from .lib_dataset import (
    TextDataset,
    prepare_dataset_llama2,
    generate_from_dataset_llama2,
)

from .lib_llm_inference import (
    invoice_to_chatgpt3_5,
    invoice_to_hf_llama,
)

from .lib_model import (
  load_peft_model_with_adpaters,
  print_trainable_parameters,
)

from .lib_pdf_manipulation import (
    convert_pdf_to_images,
    extract_text_from_pdf, 
    extract_text_from_imgs, 
)

from .lib_prompts import (
    create_llama2_prompt,
    create_openai_prompt,
)

from .lib_settings import (
    BasicInferenceRequest,
    ArtifactNames,
    FineTuningSettings,
)

from .lib_tokenizer import (
    get_tokenizer_for_model, 
    update_model_with_tokenizer,
)


## Project Level Constants 

VERSION = 1 

# Used for prompt/request data 
class BasicInferenceRequest:
  def __init__(self, token="", inference_url="", system_prompt="", content_prompt="", output_prompt=""):
    self.token = token
    self.inference_url = inference_url
    self.system_prompt = system_prompt
    self.content_prompt = content_prompt
    self.output_prompt = output_prompt

  def __str__(self):
    return f"BIR [ token: {self.token}, inference_url: {self.inference_url}, system_prompt: {self.system_prompt}, content_prompt: {self.content_prompt}, output_prompt: {self.output_prompt} ]"

  def __repr__(self):
    return f"BasicInferenceRequest(\'{self.token}\', \'{self.inference_url}\', \'{self.system_prompt}\', \'{self.content_prompt}\', \'{self.output_prompt}\')"


# Used for metadata for Fine Tuning 
class ArtifactNames:
  def __init__(self, base_model, new_model_name, repo_name, dataset):
    self.base_model = base_model
    base_model_name = base_model.split("/")[-1]
    self.adapter_model = repo_name + "/" + base_model_name + "_" + new_model_name
    self.new_model = repo_name + "/" + new_model_name 
    self.dataset = dataset 
  
  def __str__(self):  
    return f"ArtifactNames [ base_model: {self.base_model}, adapter_model: {self.adapter_model}, new_model: {self.new_model}, dataset: {self.dataset} ]"

  def __repr__(self):  
    return f"ArtifactNames(\'{self.base_model}\', \'{self.adapter_model}\', \'{self.new_model}\', \'{self.dataset}\')"


class FineTuningSettings:
  def __init__(self, base_model, new_model_name, repo_name, dataset, cache_dir="/cache"):
    self.base_model = base_model
    base_model_name = base_model.split("/")[-1]
    self.adapter_model = repo_name + "/" + base_model_name + "_" + new_model_name
    self.new_model = repo_name + "/" + new_model_name
    self.dataset = dataset
    self.cache_dir = cache_dir


# Library of Prompt related functions for various types of models

def debugPrintMessage(prompt, message = ""):
  print(f" == debug {message} ====================")
  print(prompt)
  print(" ========================================")


def createLlama2Prompt(system_prompt: str, content_prompt: str, output_prompt: str, debug = False) -> str:
  return create_llama2_prompt(system_prompt, content_prompt, output_prompt, debug = False) 

def create_llama2_prompt(system_prompt: str, content_prompt: str, output_prompt: str, debug = False) -> str:
  prompt = f"""
  <s>[INST] <<SYS>>
  {system_prompt}
  <</SYS>>

  {content_prompt} {output_prompt}[/INST]
  """.strip()

  if debug == True:
    debugPrintMessage(prompt, "prompt")

  return prompt


def create_openai_prompt(system_prompt: str, content_prompt: str, debug = False) -> str:
  prompt = f"""
  {system_prompt} 

  {content_prompt}
  """.strip()

  if debug == True:
    debugPrintMessage(prompt, "prompt")

  return prompt




from transformers import AutoTokenizer

def get_tokenizer_for_model(model, cache_dir, debug=False):

  tokenizer = AutoTokenizer(model, cache_dir)
  tokenizer.padding_side='right'

  # Add the pad token if missing the tokenizer vocabulary
  if '|<pad>|' not in tokenizer.get_vocab():
    tokenizer.add_tokens(['|<pad>|'])

  # Set the pad token
  tokenizer.pad_token = '|<pad>|'

  if debug == True:
    print("EOS token:", tokenizer.eos_token, ", id:", tokenizer.eos_token_id)
    print("Pad token: ", tokenizer.pad_token, ", ID: ", tokenizer.pad_token_id)

  return tokenizer 

def update_model_with_tokenizer(model, tokenizer, debug=False):

  # Resize token embeddings
  model.resize_token_embeddings(len(tokenizer))

  # Update pad token id in model and its config
  model.pad_token_id = tokenizer.pad_token_id
  model.config.pad_token_id = tokenizer.pad_token_id

  # Check if they are equal
  assert model.pad_token_id == tokenizer.pad_token_id, "The model's pad token ID does not match the tokenizer's pad token ID!"

  if debug == True:
    # Print the pad token ids
    print('Tokenizer pad token ID:', tokenizer.pad_token_id)
    print('Model pad token ID:', model.pad_token_id)
    print('Model config pad token ID:', model.config.pad_token_id)

    # model configuration
    print(model.config)

    # Sample string
    sample_string = ['[INST]']

    # Tokenize the stringified JSON object
    # encoded_sample = tokenizer(sample_string, truncation=True, padding=True, max_length=1024, return_tensors='pt', add_special_tokens=True)
    encoded_sample = tokenizer(sample_string, truncation=True, padding=True, max_length=1024, return_tensors='pt', add_special_tokens=False)

    # Count the number of tokens
    token_count = len(encoded_sample)

    BOS_token_id = tokenizer.bos_token_id
    EOS_token_id = tokenizer.eos_token_id
    BOS_token = tokenizer.decode([BOS_token_id])
    EOS_token = tokenizer.decode([EOS_token_id])

    print(f"Beginning of the sequence: {sample_string[0]} (BOS token: {BOS_token}, id: {BOS_token_id})")
    print(f"End of the sequence: {sample_string[-1]} (EOS token: {EOS_token}, id: {EOS_token_id})")

    print(f"The number of tokens in the string is: {token_count}")
    print(f"The ids are: {encoded_sample}")

    # Decode the input_ids
    decoded_sample = tokenizer.decode(encoded_sample['input_ids'][0], skip_special_tokens=False)

    print(f"The decoded string is: {decoded_sample}")
    print(f"The attention mask is: {encoded_sample['attention_mask']}")

  return model 


from datasets import load_dataset
import re
import torch
from torch.utils.data import Dataset

class TextDataset(Dataset):
    def __init__(self, encodings, response_lengths, input_lengths):
        self.encodings = encodings
        self.response_lengths = response_lengths
        self.input_lengths = input_lengths

    def __getitem__(self, idx):
        item = {key: val[idx].clone().detach() for key, val in self.encodings.items()}

        # Set labels to be the same as input_ids
        item["labels"] = item["input_ids"].clone()

        # Calculate the start and end positions of the response
        response_start_position = self.input_lengths[idx]
        response_end_position = self.input_lengths[idx] + self.response_lengths[idx]

        # Create a loss mask that covers only the response tokens
        item["loss_mask"] = torch.zeros_like(item["input_ids"])
        item["loss_mask"][response_start_position:response_end_position] = 1

        # Shift the loss mask to the left by one position
        shifted_loss_mask = torch.cat([item["loss_mask"][1:], torch.tensor([0])])
        item["loss_mask"] = shifted_loss_mask

        # Shift the labels to the left by one position
        item["labels"][:-1] = item["input_ids"][1:]

        # Replace the token after the response with an EOS token
        item["labels"][response_end_position - 1] = 2

        # Replace the token after the response with an 1 in the loss mask
        item["loss_mask"][response_end_position - 1] = 1

        return item

    def __len__(self):
        return len(self.encodings["input_ids"])


def prepare_dataset_llama2(dataset, tokenizer):
    # Define the roles and markers
    B_FUNC, E_FUNC = "<FUNCTIONS>", "</FUNCTIONS>"
    B_USER, E_USER = "<USER>", "</USER>"
    B_SYS, E_SYS = "<SYS>", "</SYS>"
    B_INST, E_INST = "[INST]", "[/INST]"

    # Create the formatted text with the correct roles for each part of the dialogue
    formatted_dataset = dataset.map(
        lambda x: {
            "input_text": "".join([
                f"{B_SYS}{x['systemPrompt'].strip()}{E_SYS}",
                f"{B_INST} {x['userPrompt'].strip()} {E_INST}\n\n",
                f"{x['assistantResponse'].strip()}",  # appending the EOS token in TextData...
            ]),
            "response_text": "".join([
                f"{x['assistantResponse'].strip()}",  # appending the EOS token in TextData...
            ]),
        }
    )

    # Tokenize the datasets
    encodings = tokenizer([dialogue["input_text"] for dialogue in formatted_dataset], truncation=True, padding=True, max_length=1024, return_tensors='pt', add_special_tokens=True)

    # Tokenize the response one by one without padding and special tokens for the purpose of calculating length
    response_lengths = [len(tokenizer.encode(dialogue["response_text"], truncation=True, max_length=1024, padding=False, add_special_tokens=False)) for dialogue in formatted_dataset]

    # Tokenize the input one by one without padding and with the initial special token for the purpose of calculating length
    total_lengths = [len(tokenizer.encode(dialogue["input_text"], truncation=True, max_length=1024, padding=False, add_special_tokens=True)) for dialogue in formatted_dataset]
    input_lengths = [total_length - response_length for total_length, response_length in zip(total_lengths, response_lengths)]

    # Create TextDataset
    text_dataset = TextDataset(encodings, response_lengths, input_lengths)

    return text_dataset

def generate_from_dataset_llama2(index, data, model, tokenizer):
    # Define the roles and markers
    B_INST, E_INST = "[INST]", "[/INST]"
    B_SYS, E_SYS = "<SYS>", "</SYS>\n\n"

    # Format your prompt template
    system_prompt = data['train'][index]['systemPrompt']
    user_prompt = data['train'][index]['userPrompt']
    correct_answer = data['train'][index]['assistantResponse']

    # Format your prompt template
    prompt = f"{B_SYS}{system_prompt.strip()}{E_SYS}{B_INST} {user_prompt.strip()} {E_INST}\n\n"
    print(f"Prompt: {prompt}")

    encoding = tokenizer(prompt, return_tensors="pt").to("cuda:0")
    output = model.generate(input_ids=encoding.input_ids,
                            attention_mask=encoding.attention_mask,
                            max_new_tokens=2000,
                            do_sample=True,
                            temperature=0.01,
                            eos_token_id=tokenizer.eos_token_id,
                            top_k=0)

    print()

    # Subtract the length of input_ids from output to get only the model's response
    output_text = tokenizer.decode(output[0, len(encoding.input_ids[0]):], skip_special_tokens=False)
    output_text = re.sub('\n+', '\n', output_text)  # remove excessive newline characters

    print("Generated Assistant Response: ", output_text, "\n")
    print("Correct Assistant Response:", correct_answer, "\n")